#include <wchar.h>
#include "cstrvec.h"
#include "memmem.h"
#if defined(_MSC_VER)
#  pragma warning(push)
#  pragma warning(disable : 4996) // MS reports fopen and _wfopen to be deprecated
#endif

// NOTE: MinGW/GCC may report an overwhelming number of warnings that format %z is not supported on Windows.
// Recent Windows 10 versions support it though. Specify compiler option -Wno-format to get rid of these warnings.
// Macros PRIuPTR (and PRIdPTR, respectively) defined in <inttypes.h> are reasonable to work around the missing
// %zu and %zd format specifiers on older Windows versions.

static void cs_init_rdfile_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  puts("Enter more than 5 characters:");
  cstr cs = cs_init_rdfile(stdin, 5);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_init_rdline_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  puts("That's the rest of the previous input:");
  cstr cs = cs_init_rdline(stdin);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_init_rdentirefile_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const char *fname = "test.~tmp";

  {
    const char fcontent[] = "This is a test,\nto show you\nhow to read\nthe whole file.";
    FILE *stream = fopen(fname, "wb");
    if (stream)
    {
      fwrite(fcontent, 1U, sizeof(fcontent) - 1, stream);
      fclose(stream);
    }
  }

  FILE *stream = fopen(fname, "rb"); // Open for reading.
  if (stream)
  {
    cstr cs = cs_init_rdentirefile(stream);
    fclose(stream);
    remove(fname); // The file isn't used anymore.
    printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
    cs_free(cs);
  }
}

static void cs_init_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("aaa");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa"
               "aaaaaaaaaa");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_init_format_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init_format("format int %d", -123456);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init_format("format no arguments");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_init_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init_n("aaax", 3);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init_n("aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa",
                 300);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_max_size_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);
  printf("cstr size limit: %zu\n\n", cs_max_size());
}

static void cs_append_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("xxx");
  cs_append(&cs, "aaa");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("xxx");
  cs_append(&cs, "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_append_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("xxx");
  cs_append_n(&cs, "aaax", 3);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("xxx");
  cs_append_n(&cs, "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa",
              300);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_assign_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("");
  cs_assign(&cs, "aaa");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_assign(&cs, "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa"
                 "aaaaaaaaaa");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_assign_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("");
  cs_assign_n(&cs, "aaax", 3);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_assign_n(&cs, "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa"
                   "aaaaaaaaaa",
              300);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_at_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("ABC");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n%d\n\n", cs_capacity(cs), cs_size(cs), cs, (int)cs_at(cs, 2));
  printf("capacity %zu\nlength   %zu\n\"%s\"\n%d\n\n", cs_capacity(cs), cs_size(cs), cs, (int)cs_at(cs, 100));
  cs_free(cs);
}

static void cs_clear_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("aaa");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n%s\n\n", cs_capacity(cs), cs_size(cs), cs, cs_empty(cs) ? "true" : "false");
  cs_clear(cs);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n%s\n\n", cs_capacity(cs), cs_size(cs), cs, cs_empty(cs) ? "true" : "false");
  cs_free(cs);
}

static void cs_erase_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  cs_erase(cs, 1, 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_erase(cs, 3, 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_erase(cs, 10, 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_fix_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  cs_fix(&cs, 4, "xy", 1 | 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_assign(&cs, "abcd");
  cs_fix(&cs, 11, "xy", 1);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_assign(&cs, "abcd");
  cs_fix(&cs, 11, "xy", 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_assign(&cs, "abcd");
  cs_fix(&cs, 11, "xy", 1 | 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_assign(&cs, "abcd");
  cs_fix(&cs, 2, NULL, 1);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_assign(&cs, "abcd");
  cs_fix(&cs, 2, NULL, 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_assign(&cs, "abcd");
  cs_fix(&cs, 2, NULL, 1 | 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_free(cs);
}

static void cs_insert_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  cs_insert(&cs, 2, "xxx");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_insert(&cs, 10, "xxx");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_insert_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  cs_insert_n(&cs, 2, "xxxyz", 3);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_insert_n(&cs, 10, "xxxyz", 3);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_pop_back_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("a");
  char ch = cs_pop_back(cs);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n%d\n\n", cs_capacity(cs), cs_size(cs), cs, (int)ch);

  ch = cs_pop_back(cs);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n%d\n\n", cs_capacity(cs), cs_size(cs), cs, (int)ch);
  cs_free(cs);
}

static void cs_push_back_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("");
  cs_push_back(&cs, 'a');
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_push_back(&cs, 'b');
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_replace_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  cs_replace(&cs, 1, 2, "x");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_replace(&cs, 1, 2, "xx");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_replace(&cs, 1, 2, "xxx");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_replace(&cs, 3, 2, "xxx");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_replace(&cs, 10, 2, "xxx");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_replace_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  cs_replace_n(&cs, 1, 2, "xyz", 1);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_replace_n(&cs, 1, 2, "xxyz", 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_replace_n(&cs, 1, 2, "xxxyz", 3);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_replace_n(&cs, 3, 2, "xxxyz", 3);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_init("abcd");
  cs_replace_n(&cs, 10, 2, "xxxyz", 3);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_reserve_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_reserve(&cs, 500);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_resize_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  cs_resize(&cs, 500, 'x');
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_resize(&cs, 5, 'z');
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_resize(&cs, 2, 'z');
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_reverse_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  cs_reverse(cs);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_free(cs);
}

static void cs_shrink_to_fit_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs = cs_shrink_to_fit(&cs);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_substr_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr original = cs_init("abcd");

  cstr cs = cs_substr(original, 2, 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_substr(original, 3, 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs = cs_substr(original, 4, 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);

  cs_free(original);
}

static void cs_trim_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);
  cstr cs = cs_init("yxyxyxABCyxyxyx");
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_trim(cs, "xy", 1 | 2);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);

  cs_free(cs);
}

static void cs_update_length_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init_zero(300);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  memcpy(cs, "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa"
             "aaaaaaaaaa",
         301);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_update_length(cs, 300);
  printf("capacity %zu\nlength   %zu\n\"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static void cs_utf8_handle_partials_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  u8state_t state = { 0 };
  cstr cs = cs_init("\x7E" // Tilde
                    "\xC3\xB6" // Umlaut oe
                    "\xE2\x82" // Euro Symbol *** incomplete *** first two bytes
  );
  cs_utf8_handle_partials(&cs, &state);
  printf("length   %zu\ncount    %zu\n\n", cs_size(cs), cs_utf8_count(cs));
  cs_free(cs);

  cs = cs_init("\xAC" // Euro Symbol *** incomplete *** last byte
               "\xF0\xA4\xBD\x9C" // CJK Ideogramm
  );
  cs_utf8_handle_partials(&cs, &state);
  printf("length   %zu\ncount    %zu\n\n", cs_size(cs), cs_utf8_count(cs));
  cs_free(cs);
}

static void cs_utf8_count_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("\xEF\xBF\xBD" // UTF-8 Byte Order Mark
                    "\x7E" // Tilde
                    "\xC3\xB6" // Umlaut oe
                    "\xE0\x82\xAC" // **** INVALID ****
                    "\xF0\xA4\xBD\x9C" // CJK Ideogramm
  );
  printf("length   %zu\ncount    %zu\n\n", cs_size(cs), cs_utf8_count(cs));
  cs_free(cs);

  cs = cs_init("\xEF\xBF\xBD" // UTF-8 Byte Order Mark
               "\x7E" // Tilde
               "\xC3\xB6" // Umlaut oe
               "\xE2\x82\xAC" // Euro Symbol
               "\xF0\xA4\xBD\x9C" // CJK Ideogramm
  );
  printf("length   %zu\ncount    %zu\n\n", cs_size(cs), cs_utf8_count(cs));
  cs_free(cs);
}

static void cs_utf8_hasbom_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("\x7E" // Tilde
                    "\xC3\xB6" // Umlaut oe
                    "\xE2\x82\xAC" // Euro Symbol
                    "\xF0\xA4\xBD\x9C" // CJK Ideogramm
  );
  printf("%s\n\n", cs_utf8_hasbom(cs) ? "true" : "false");
  cs_free(cs);

  cs = cs_init("\xEF\xBF\xBD" // UTF-8 Byte Order Mark
               "\x7E" // Tilde
               "\xC3\xB6" // Umlaut oe
               "\xE2\x82\xAC" // Euro Symbol
               "\xF0\xA4\xBD\x9C" // CJK Ideogramm
  );
  printf("%s\n\n", cs_utf8_hasbom(cs) ? "true" : "false");
  cs_free(cs);
}

static void cs_utf8_idx_of_nth_codepoint_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("\x7E" // Tilde
                    "\xC3\xB6" // Umlaut oe
                    "\xE2\x82\xAC" // Euro Symbol
                    "\xF0\xA4\xBD\x9C" // CJK Ideogramm
  );
  printf("index %zu\n\n", cs_utf8_idx_of_nth_codepoint(cs, 3));
  cs_free(cs);
}

static void cs_utf8_replace_invalids_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("\x7E" // Tilde
                    "\xC3\xB6" // Umlaut oe
                    "\xE0\x82\xAC" // **** INVALID **** (will result in 3 replacement characters with 3 code units each)
                    "\xF0\xA4\xBD\x9C" // CJK Ideogramm
  );
  cs_utf8_replace_invalids(&cs, NULL);
  printf("length   %zu\ncount    %zu\n\n", cs_size(cs), cs_utf8_count(cs));
  cs_free(cs);
}

//////////////////////////////////////////////////////////////////////////////////////

#if defined(_WIN32)

static void csw_init_rdfile_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  puts("Enter more than 5 characters:");
  cstrw cs = csw_init_rdfile(stdin, 5);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_init_rdline_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  puts("That's the rest of the previous input:");
  cstrw cs = csw_init_rdline(stdin);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_init_rdentirefile_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const wchar_t *fname = L"test.~tmp";

  {
    const wchar_t fcontent[] = L"This is a test,\nto show you\nhow to read\nthe whole file.";
    FILE *stream = _wfopen(fname, L"wb");
    if (stream)
    {
      fwrite(fcontent, sizeof(wchar_t), sizeof(fcontent) / sizeof(wchar_t) - 1, stream);
      fclose(stream);
    }
  }

  FILE *stream = _wfopen(fname, L"rb"); // Open for reading.
  if (stream)
  {
    cstrw csw = csw_init_rdentirefile(stream);
    fclose(stream);
    _wremove(fname); // The file isn't used anymore.
    wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(csw), csw_size(csw), csw);
    csw_free(csw);
  }
}

static void csw_init_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"aaa");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_init_format_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init_format(L"format int %d", -123456);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_init_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init_n(L"aaax", 3);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init_n(L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
                  300);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_max_size_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);
  printf("cstrw size limit: %zu\n\n", csw_max_size());
}

static void csw_append_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"xxx");
  csw_append(&cs, L"aaa");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"xxx");
  csw_append(&cs, L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_append_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"xxx");
  csw_append_n(&cs, L"aaax", 3);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"xxx");
  csw_append_n(&cs, L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                    L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                    L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                    L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                    L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
               300);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_assign_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"");
  csw_assign(&cs, L"aaa");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_assign(&cs, L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                  L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_assign_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"");
  csw_assign_n(&cs, L"aaax", 3);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_assign_n(&cs, L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                    L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                    L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                    L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
                    L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
               300);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_at_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"ABC");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n%d\n\n", csw_capacity(cs), csw_size(cs), cs, (int)csw_at(cs, 2));
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n%d\n\n", csw_capacity(cs), csw_size(cs), cs, (int)csw_at(cs, 100));
  csw_free(cs);
}

static void csw_clear_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"aaa");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n%s\n\n", csw_capacity(cs), csw_size(cs), cs, csw_empty(cs) ? L"true" : L"false");
  csw_clear(cs);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n%s\n\n", csw_capacity(cs), csw_size(cs), cs, csw_empty(cs) ? L"true" : L"false");
  csw_free(cs);
}

static void csw_erase_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  csw_erase(cs, 1, 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_erase(cs, 3, 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_erase(cs, 10, 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_fix_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  csw_fix(&cs, 4, L"xy", 1 | 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_assign(&cs, L"abcd");
  csw_fix(&cs, 11, L"xy", 1);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_assign(&cs, L"abcd");
  csw_fix(&cs, 11, L"xy", 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_assign(&cs, L"abcd");
  csw_fix(&cs, 11, L"xy", 1 | 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_assign(&cs, L"abcd");
  csw_fix(&cs, 2, NULL, 1);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_assign(&cs, L"abcd");
  csw_fix(&cs, 2, NULL, 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_assign(&cs, L"abcd");
  csw_fix(&cs, 2, NULL, 1 | 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_free(cs);
}

static void csw_insert_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  csw_insert(&cs, 2, L"xxx");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_insert(&cs, 10, L"xxx");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_insert_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  csw_insert_n(&cs, 2, L"xxxyz", 3);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_insert_n(&cs, 10, L"xxxyz", 3);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_pop_back_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"a");
  wchar_t ch = csw_pop_back(cs);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n%d\n\n", csw_capacity(cs), csw_size(cs), cs, (int)ch);

  ch = csw_pop_back(cs);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n%d\n\n", csw_capacity(cs), csw_size(cs), cs, (int)ch);
  csw_free(cs);
}

static void csw_push_back_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"");
  csw_push_back(&cs, L'a');
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_push_back(&cs, L'b');
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_replace_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  csw_replace(&cs, 1, 2, L"x");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_replace(&cs, 1, 2, L"xx");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_replace(&cs, 1, 2, L"xxx");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_replace(&cs, 3, 2, L"xxx");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_replace(&cs, 10, 2, L"xxx");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_replace_n_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  csw_replace_n(&cs, 1, 2, L"xyz", 1);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_replace_n(&cs, 1, 2, L"xxyz", 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_replace_n(&cs, 1, 2, L"xxxyz", 3);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_replace_n(&cs, 3, 2, L"xxxyz", 3);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_init(L"abcd");
  csw_replace_n(&cs, 10, 2, L"xxxyz", 3);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_reserve_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_reserve(&cs, 500);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_resize_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  csw_resize(&cs, 500, L'x');
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_resize(&cs, 5, L'z');
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_resize(&cs, 2, L'z');
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_reverse_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  csw_reverse(cs);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_free(cs);
}

static void csw_shrink_to_fit_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init(L"abcd");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  cs = csw_shrink_to_fit(&cs);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

static void csw_substr_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw original = csw_init(L"abcd");

  cstrw cs = csw_substr(original, 2, 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_substr(original, 3, 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  cs = csw_substr(original, 4, 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);

  csw_free(original);
}

static void csw_trim_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);
  cstrw cs = csw_init(L"yxyxyxABCyxyxyx");
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_trim(cs, L"xy", 1 | 2);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);

  csw_free(cs);
}

static void csw_update_length_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstrw cs = csw_init_zero(300);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  memcpy(cs, L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
             L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
             L"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
         301 * sizeof(wchar_t));
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_update_length(cs, 300);
  wprintf(L"capacity %zu\nlength   %zu\n\"%s\"\n\n", csw_capacity(cs), csw_size(cs), cs);
  csw_free(cs);
}

#endif

//////////////////////////////////////////////////////////////////////////////////////

static void memmem_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  ccstr cs = cs_init_n("abcdefgha\0zbcdefa\0bcdefg", 24);
  const char *found = (const char *)memmem(cs, 24, "a\0b", 3);
  printf("offset: %zd\n\n", (found - cs));
  cs_free(cs);
}

static void memrmem_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  ccstr cs = cs_init_n("abcdefgha\0bcdefga\0zbcdef", 24);
  const char *found = (const char *)memrmem(cs, 24, "a\0b", 3);
  printf("offset: %zd\n\n", (found - cs));
  cs_free(cs);
}

static void strrstr_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const char *str = "abcdefghaxcdefghaycdefgh";
  const char *found = strrstr(str, "axc");
  printf("offset: %zd\n\n", (found - str));
}

static void wcsrstr_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const wchar_t *str = L"abcdefghaxcdefghaycdefgh";
  const wchar_t *found = wcsrstr(str, L"axc");
  printf("offset: %zd\n\n", (found - str));
}

//////////////////////////////////////////////////////////////////////////////////////

static void cv_init_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const int arr[] = { 1, 2, 3, 4 };
  cvec cv = cv_init(arr, 4, sizeof(int));

  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_free(cv);
}

static void cv_init_zero_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cvec cv = cv_init_zero(4, sizeof(int));
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));

  puts("\n");
  cv_free(cv);
}

static void cv_max_size_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);
  printf("cvec size limit for int: %zu\n\n", cv_max_size(sizeof(int)));
}

static void cv_append_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const int arr[] = { 1, 2, 3, 4 };
  cvec cv = cv_init(arr, 4, sizeof(int));

  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_append(&cv, arr, 4);
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_free(cv);
}

static void cv_assign_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const int arr[] = { 1, 2, 3, 4, 5, 6 };
  const int arr2[] = { 1, 2, 3, 4 };
  cvec cv = cv_init(arr, 6, sizeof(int));
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_assign(&cv, arr2, 4);
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_free(cv);
}

static void cv_at_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const int arr[] = { 1, 2, 3, 4 };
  cvec cv = cv_init(arr, 4, sizeof(int));

  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  printf(" %d\n\n", *PTRTO(cv_at(cv, 3), int));

  cv_free(cv);
}

static void cv_clear_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const int arr[] = { 1, 2, 3, 4 };
  cvec cv = cv_init(arr, 4, sizeof(int));
  printf("capacity %zu\nsize     %zu\n%s\n\n", cv_capacity(cv), cv_size(cv), cv_empty(cv) ? "true" : "false");

  cv_clear(cv);
  printf("capacity %zu\nsize     %zu\n%s\n\n", cv_capacity(cv), cv_size(cv), cv_empty(cv) ? "true" : "false");

  cv_free(cv);
}

static void cv_erase_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const int arr[] = { 1, 2, 3, 4 };
  cvec cv = cv_init(arr, 4, sizeof(int));

  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_erase(cv, 1, 2);
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_free(cv);
}

static bool for_each_int_callback(void *p_data, void *user_parameter)
{
  const int dat = *(int *)p_data;
  const int max = *(int *)user_parameter;
  if (dat > max)
    return false;

  printf(" %d\n", dat);
  return true;
}

static void cv_for_each_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const int arr[] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 };
  int max = 5;
  cvec cv = cv_init(arr, 10, sizeof(int));
  cv_for_each(cv, 1, (size_t)-1, for_each_int_callback, &max);
  cv_free(cv);
}

static void cv_insert_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const int arr[] = { 1, 2, 3, 4 };
  cvec cv = cv_init(arr, 4, sizeof(int));

  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_insert(&cv, 2, arr, 4);
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_free(cv);
}

static void cv_pop_back_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const int arr[] = { 1, 2, 3, 4 };
  cvec cv = cv_init(arr, 4, sizeof(int));

  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  printf(" %d\n", *PTRTO(cv_pop_back(cv), int));
  printf("capacity %zu\nsize     %zu\n\n", cv_capacity(cv), cv_size(cv));

  cv_free(cv);
}

static void cv_push_back_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cvec cv = cv_init(MAKEPTR(int, 1, 2, 3, 4), 4, sizeof(int));

  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_push_back(&cv, MAKEPTR(int, 5));
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_free(cv);
}

static void cv_reserve_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cvec cv = cv_init(MAKEPTR(int, 1, 2, 3, 4), 4, sizeof(int));
  printf("capacity %zu\nsize     %zu\n\n", cv_capacity(cv), cv_size(cv));

  cv_reserve(&cv, 100);
  printf("capacity %zu\nsize     %zu\n\n", cv_capacity(cv), cv_size(cv));

  cv_free(cv);
}

static void cv_resize_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cvec cv = cv_init(MAKEPTR(int, 1, 2, 3, 4), 4, sizeof(int));
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_resize(&cv, 100, MAKEPTR(int, 5));
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_resize(&cv, 2, NULL);
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_shrink_to_fit(&cv);
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_free(cv);
}

static void cv_update_size_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cvec cv = cv_init_zero(2, sizeof(int));
  printf("capacity %zu\nsize     %zu\n\n", cv_capacity(cv), cv_size(cv));

  PTRTO(cv, int)
  [0] = 1;

  PTRTO(cv, int)
  [1] = 2;

  printf("capacity %zu\nsize     %zu\n\n", cv_capacity(cv), cv_size(cv));

  cv_update_size(cv, 2);
  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  for (size_t i = 0; i < cv_size(cv); ++i)
    printf(" %d", PTRTO(cv, int)[i]);

  puts("\n");
  cv_free(cv);
}

static void cvec_of_cvec_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  // cvec of cvec
  cvec multi = cv_init_zero(2, sizeof(cvec));

  cv_push_back(&multi, MAKEPTR(cvec, cv_init(MAKEPTR(int, 1, 2, 3, 4, 5, 6), 6, sizeof(int))));
  cv_push_back(&multi, MAKEPTR(cvec, cv_init(MAKEPTR(int, 1, 2, 3, 4), 4, sizeof(int))));

  printf("capacity %zu\nsize     %zu\n-------------\n\n", cv_capacity(multi), cv_size(multi));

  // restrict the scope of tmp_cv to the braced blocks each
  {
    const ccvec tmp_cv = *PTRTO(cv_at(multi, 0), cvec);
    printf("capacity %zu\nsize     %zu\n", cv_capacity(tmp_cv), cv_size(tmp_cv));
    for (size_t i = 0; i < cv_size(tmp_cv); ++i)
      printf(" %d", PTRTO(tmp_cv, int)[i]);

    puts("\n");
  }
  {
    const ccvec tmp_cv = *PTRTO(cv_at(multi, 1), cvec);
    printf("capacity %zu\nsize     %zu\n", cv_capacity(tmp_cv), cv_size(tmp_cv));
    for (size_t i = 0; i < cv_size(tmp_cv); ++i)
      printf(" %d", PTRTO(tmp_cv, int)[i]);

    puts("\n");
  }

  for (size_t idx = 0; idx < cv_size(multi); ++idx)
    cv_free(*PTRTO(cv_at(multi, idx), cvec));

  cv_free(multi);
}

//////////////////////////////////////////////////////////////////////////////////////

static void cs_join_cs_split_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  // create a cvec of cstr
  cvec arrofcs = cv_init(MAKECSTRPTR(cs_init("abc"), cs_init("defghi"), cs_init("jklm")), 3, sizeof(cstr));

  // join it in a cstr and print what we have
  cstr joined = cs_join(arrofcs, ", ");
  printf("capacity %zu\nsize     %zu\n \"%s\"\n\n", cs_capacity(joined), cs_size(joined), joined);

  // destroy the old cvec of cstr
  cstrvec_free(arrofcs);

  /////////////////////////////////////

  // split the cstr into a cvec
  arrofcs = cs_split(joined, ", ", (size_t)-1);

  // we don't need the joined cstr anymore, destroy it
  cs_free(joined);

  // print the properties of the cvec
  printf("capacity %zu\nsize     %zu\n", cv_capacity(arrofcs), cv_size(arrofcs));
  // and the containing cstr tokens
  for (size_t idx = 0; idx < cv_size(arrofcs); ++idx)
    printf(" \"%s\"\n", *PTRTOCSTR(cv_at(arrofcs, idx)));

  // again destroy the cvec of cstr
  cstrvec_free(arrofcs);
}

static bool print_byte_callback(void *p_element, void *unused)
{
  (void)unused;
  printf(" \\x%02X", *(unsigned char *)p_element);
  return true;
}

static void cvbyte_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  cstr cs = cs_init("abcd");

  cvec cv = cvbyte_from_cs(cs);
  cs_free(cs);

  printf("capacity %zu\nsize     %zu\n", cv_capacity(cv), cv_size(cv));
  cv_for_each(cv, 0, (size_t)-1, print_byte_callback, NULL);
  puts("\n");

  cs = cs_from_cvbyte(cv);
  cv_free(cv);

  printf("capacity %zu\nlength   %zu\n \"%s\"\n\n", cs_capacity(cs), cs_size(cs), cs);
  cs_free(cs);
}

static bool print_cstr_callback(void *p_element, void *p_param)
{
  printf("%3u\t%s\n", ++*(unsigned *)p_param, *(cstr *)p_element);
  return true;
}

static void cstrvec_rdlines_TEST(void)
{
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n*** %s ***\n\n", __func__);

  const char *fname = "test.~tmp";

  {
    const char fcontent[] = "This is a test,\nto show you\nhow a file\nis read line-wise.\n\nLast line.";
    FILE *stream = fopen(fname, "w");
    if (stream)
    {
      fwrite(fcontent, 1U, sizeof(fcontent) - 1, stream);
      fclose(stream);
    }
  }

  FILE *stream = fopen(fname, "r"); // Open for reading.
  if (stream)
  {
    cvec lines = cstrvec_rdlines(stream, (size_t)-1); // Create a vector of strings (lines) of the file.

    fclose(stream); // File has been processed.

    // For the output we want to prepend a line number which will be incremented in the callback function.
    unsigned cnt = 0U;
    // The callback function prints every string in the vector in a new line.
    cv_for_each(lines, 0U, (size_t)-1, print_cstr_callback, &cnt);

    cstrvec_free(lines); // Release the memory of the strings and the vector container.
  }

  remove(fname); // The file isn't used anymore.
}

#if defined(_MSC_VER)
#  pragma warning(pop) // C4996
#endif

int main(void)
{
  cs_init_rdfile_TEST();
  cs_init_rdline_TEST();
  cs_init_rdentirefile_TEST();
  cs_init_TEST();
  cs_init_format_TEST();
  cs_init_n_TEST();
  cs_max_size_TEST();
  cs_append_TEST();
  cs_append_n_TEST();
  cs_assign_TEST();
  cs_assign_n_TEST();
  cs_at_TEST();
  cs_clear_TEST(); // + cs_empty
  cs_erase_TEST();
  cs_fix_TEST();
  cs_insert_TEST();
  cs_insert_n_TEST();
  cs_pop_back_TEST();
  cs_push_back_TEST();
  cs_replace_TEST();
  cs_replace_n_TEST();
  cs_reserve_TEST();
  cs_resize_TEST();
  cs_reverse_TEST();
  cs_shrink_to_fit_TEST();
  cs_substr_TEST();
  cs_trim_TEST();
  cs_update_length_TEST(); // + cs_init_zero
  cs_utf8_handle_partials_TEST();
  cs_utf8_count_TEST();
  cs_utf8_hasbom_TEST();
  cs_utf8_idx_of_nth_codepoint_TEST();
  cs_utf8_replace_invalids_TEST();

  // Test functions for wide string will also print them but only the Windows console supports a mixed output of char- and
  // wchar_t-based strings. Terminals require a conditioning and reopening of the stdout.
#if defined(_WIN32)
  csw_init_rdfile_TEST();
  csw_init_rdline_TEST();
  csw_init_rdentirefile_TEST();
  csw_init_TEST();
  csw_init_format_TEST();
  csw_init_n_TEST();
  csw_max_size_TEST();
  csw_append_TEST();
  csw_append_n_TEST();
  csw_assign_TEST();
  csw_assign_n_TEST();
  csw_at_TEST();
  csw_clear_TEST(); // + csw_empty
  csw_erase_TEST();
  csw_fix_TEST();
  csw_insert_TEST();
  csw_insert_n_TEST();
  csw_pop_back_TEST();
  csw_push_back_TEST();
  csw_replace_TEST();
  csw_replace_n_TEST();
  csw_reserve_TEST();
  csw_resize_TEST();
  csw_reverse_TEST();
  csw_shrink_to_fit_TEST();
  csw_substr_TEST();
  csw_trim_TEST();
  csw_update_length_TEST(); // + csw_init_zero
#endif

  memmem_TEST();
  memrmem_TEST();
  strrstr_TEST();
  wcsrstr_TEST();

  cv_init_TEST();
  cv_init_zero_TEST();
  cv_max_size_TEST();
  cv_append_TEST();
  cv_assign_TEST();
  cv_at_TEST();
  cv_clear_TEST(); // + cs_empty
  cv_erase_TEST();
  cv_for_each_TEST();
  cv_insert_TEST();
  cv_pop_back_TEST();
  cv_push_back_TEST();
  cv_reserve_TEST();
  cv_resize_TEST(); // + cv_shrink_to_fit
  cv_update_size_TEST();
  cvec_of_cvec_TEST();

  cs_join_cs_split_TEST();
  cvbyte_TEST();
  cstrvec_rdlines_TEST();

  return 0;
}
